## Copyright 2014 Cognitect. All Rights Reserved.
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS-IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

import itertools
from .pyversion import imap, izip


def mapcat(f, i):
    return itertools.chain.from_iterable(imap(f, i))


def pairs(i):
    return izip(*[iter(i)] * 2)


cycle = itertools.cycle


def take(n, i):
    return itertools.islice(i, 0, n)
